import java.awt.*;
import java.io.*;

import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;


/**
 * Dialog class for showing help contents.
 */
public class HelpDialog extends JFrame implements ListSelectionListener{
	JTextArea ta; // contents
	JScrollPane sp; // topics
	
	private static HelpDialog hd; // singleton dialog object.
	
	/**
	 * Returns the singleton help dialog object.
	 * @param icApp Main application.
	 * @return HelpDialog object.
	 */
	public static HelpDialog getHelpDialog(Laf icApp){
		if(hd==null) hd=new HelpDialog(icApp);
		return hd;
	}
	
	/**
	 * Constructor:
	 * @param icApp Main application.
	 */
	public HelpDialog(Laf icApp){
		super("Laf Help");
		
		// list of topics.
		JList l=new JList(new String[]{"Foreground extraction", "Scissoring", "Warping", "Mosaicing"});
		sp=new JScrollPane(l);
		l.addListSelectionListener(this);
		l.setDragEnabled(false);
		l.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		getContentPane().add(sp, BorderLayout.WEST);
		
		// help contents
		ta=new JTextArea();
		ta.setEditable(false);
		ta.setLineWrap(true);
		ta.setWrapStyleWord(true);
		
		sp=new JScrollPane(ta);
		sp.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		sp.setPreferredSize(new Dimension(300, 200));
		getContentPane().add(sp);
		
		pack();
		setLocation((icApp.getWidth()-getWidth())/2, (icApp.getHeight()-getHeight())/2);
		setVisible(true);
	}

	/* (non-Javadoc)
	 * @see javax.swing.event.ListSelectionListener#valueChanged(javax.swing.event.ListSelectionEvent)
	 */
	public void valueChanged(ListSelectionEvent e) {
		if(e.getValueIsAdjusting()){
			// count only one event on list item change.
			String topic=(String)((JList)e.getSource()).getSelectedValue();
			String line, doc="";
			try{
				// read the topic file for contents.
				BufferedReader br=new BufferedReader(new FileReader("help/"+topic+".txt"));
				while((line=br.readLine())!=null) doc+="\n"+line;
				// get rid of the 1st newline.
				if(doc.length()>0) doc=doc.substring(1);
			}catch(IOException ex){
				doc="Couldn't find the contents of this topic.";
			}
			// create new text area to hold the contents.
			ta=new JTextArea(doc);
			ta.setEditable(false);
			ta.setLineWrap(true);
			ta.setWrapStyleWord(true);
			sp.getViewport().setView(ta);
			sp.setColumnHeaderView(new JLabel(topic));
		}
	}
}
