import java.awt.event.MouseEvent;

/**
 * Interface for passing events to a drawing tool.
 */
public interface ICOperation {
	
	/** Event handler for mouse presses.
	 * @param e Event object describing the mouse event.
	 * @return true to redraw component, false otherwise.
	 */
	public boolean mousePressed(MouseEvent e);
	
	/** Event handler for mouse releases.
	 * @param e Event object describing the mouse event.
	 * @return true to redraw component, false otherwise.
	 */
	public boolean mouseReleased(MouseEvent e);
	
	/** Event handler for mouse enterings.
	 * @param e Event object describing the mouse event.
	 * @return true to redraw component, false otherwise.
	 */
	public boolean mouseEntered(MouseEvent e);
	
	/** Event handler for mouse exits.
	 * @param e Event object describing the mouse event.
	 * @return true to redraw component, false otherwise.
	 */
	public boolean mouseExited(MouseEvent e);
		
	/** Event handler for mouse clicks.
	 * @param e Event object describing the mouse event.
	 * @return true to redraw component, false otherwise.
	 */
	public boolean mouseClicked(MouseEvent e);

	/** Event handler for the mouse move event.
	 * @param e Object describing the event.
	 * @return true to redraw component, false otherwise.
	 */
	public boolean mouseMoved(MouseEvent e);

	/** Event handler for the mouse drag event.
	 * @param e Object describing the event.
	 * @return true to redraw component, false otherwise.
	 */
	public boolean mouseDragged(MouseEvent e);
	
	/**
	 * Do something before the component is drawn.
	 */
	public void beforeDrawing();
}
