import java.awt.event.*;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.regex.Pattern;


/**
 * Manages and delegates keyboard events.
 */
public class KeyBoardManager implements KeyListener {
	private Object InputEvent;
	// singletop KeyBoardManager object.
	private static KeyBoardManager kbMgr=new KeyBoardManager();
	private Hashtable pressList;
	private Hashtable releaseList;
	private Hashtable hitList;
	private Hashtable allKeysDown;
	// regular expressions for different types of keys and characters.
	private Pattern fnKey=Pattern.compile("^F[0-9][0-2]?");
	private Pattern numKey=Pattern.compile("[0-9]");
	private Pattern alphaKey=Pattern.compile("[a-zA-Z]");
	private Pattern alphanumKey=Pattern.compile("[0-9a-zA-Z]");
	private Pattern nonalphanumKey=Pattern.compile("[^0-9a-zA-Z]");

	/**
	 * Constructor:
	 */
	public KeyBoardManager(){
		pressList=new Hashtable();
		releaseList=new Hashtable();
		hitList=new Hashtable();
		allKeysDown=new Hashtable();
	}
	
	/**
	 * Returns the singleton KeyBoardManager object.
	 * @return KeyBoardManager object.
	 */
	public static KeyBoardManager getKeyBoardManager(){
		if(kbMgr==null) kbMgr=new KeyBoardManager();
		return kbMgr;
	}

	/* (non-Javadoc)
	 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
	 */
	public void keyPressed(KeyEvent e) {
		String mods=KeyEvent.getKeyModifiersText(e.getModifiers());
		String key=KeyEvent.getKeyText(e.getKeyCode());
		char ch=e.getKeyChar();
		if(mods.length()>0){
			// go through modifiers.
			String [] mList=mods.split("\\+");
			mods="";
			for(int i=0; i<mList.length; i++){
				if(!mList[i].equals(key)){
					mods+=mList[i]+"+";
				}
			}
			if(mods.length()>0)
				mods=mods.substring(0, mods.length()-1)+' ';
		}
		if(e.getKeyChar()==KeyEvent.CHAR_UNDEFINED){
			// non-char key was entered.
			if(allKeysDown.get(key)==null){
				allKeysDown.put(key, mods);
			}
		}else{
			key=key.toLowerCase();
			if(allKeysDown.get(key)==null){
				allKeysDown.put(key, mods);
			}
		}
		if(mods.length()>0)
			firePressEvent(mods+key);
		firePressEvent("All");
		
		boolean [] fired=new boolean[8];
		Boolean isChar;
		// go through all pressed keys and fire events for them.
		for(Enumeration enum=allKeysDown.keys(); enum.hasMoreElements();){
			key=(String)enum.nextElement();
			firePressEvent(key);
			fireGeneralPressEvents(key, key.length()==1, fired);
		}
	}
	
	/**
	 * Fires general events for press event.
	 * @param key Key to fire events for.
	 * @param isChar true if it's a character, false o/w.
	 * @param fired Array of booleans telling whether a particular general event was fired already or not.
	 */
	private void fireGeneralPressEvents(String key, boolean isChar, boolean [] fired){
		if(!fired[0] && !isChar){
			firePressEvent("AllKeys");
			fired[0]=true;
			if(!fired[2] && fnKey.matcher(key).matches()){
				firePressEvent("AllFnKeys");
				fired[2]=true;
			}
		}else if(!fired[3] && isChar){
			firePressEvent("AllChars");
			fired[3]=true;
			if(!fired[6] && alphanumKey.matcher(key).matches()){
				firePressEvent("AllAlphaNum");
				fired[6]=true;
			}
			if(!fired[4] && numKey.matcher(key).matches()){
				firePressEvent("AllNum");
				fired[4]=true;
			}else if(!fired[5] && alphaKey.matcher(key).matches()){
				firePressEvent("AllAlpha");
				fired[5]=true;
			}else if(!fired[7] && nonalphanumKey.matcher(key).matches()){
				firePressEvent("AllNonAlphaNum");
				fired[7]=true;
			}
		}
	}
	
	/**
	 * Fires general events for hit event.
	 * @param key Key to fire events for.
	 * @param isChar true if it's a character, false o/w.
	 */
	private void fireGeneralHitEvents(String key, boolean isChar){
		fireHitEvent("All");
		if(!isChar){
			fireHitEvent("AllKeys");
			if(fnKey.matcher(key).matches()){
				fireHitEvent("AllFnKeys");
			}
		}else if(isChar){
			fireHitEvent("AllChars");
			if(alphanumKey.matcher(key).matches()){
				fireHitEvent("AllAlphaNum");
			}
			if(numKey.matcher(key).matches()){
				fireHitEvent("AllNum");
			}else if(alphaKey.matcher(key).matches()){
				fireHitEvent("AllAlpha");
			}else if(nonalphanumKey.matcher(key).matches()){
				fireHitEvent("AllNonAlphaNum");
			}
		}
	}

	/* (non-Javadoc)
	 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
	 */
	public void keyReleased(KeyEvent e) {
		String key=KeyEvent.getKeyText(e.getKeyCode());
		char ch=e.getKeyChar();
		String mods=(String)allKeysDown.remove(key);
		if(ch!=KeyEvent.CHAR_UNDEFINED){
			// the key is a character.
			key=key.toLowerCase();
			mods=(String)allKeysDown.remove(key);
			// fire egenral events.
			firePressEvent("AllChars");
			if(alphanumKey.matcher(key).matches()){
				firePressEvent("AllAlphaNum");
			}
			if(numKey.matcher(key).matches()){
				firePressEvent("AllNum");
			}else if(alphaKey.matcher(key).matches()){
				firePressEvent("AllAlpha");
			}else if(nonalphanumKey.matcher(key).matches()){
				firePressEvent("AllNonAlphaNum");
			}
		}else{
			firePressEvent("AllKeys");
			if(fnKey.matcher(key).matches()){
				firePressEvent("AllFnKeys");
			}
		}
		if(mods.length()>0){
			fireReleaseEvent(mods+key);
			fireHitEvent(mods+key);
		}
		firePressEvent(key);
		fireReleaseEvent("All");

		fireHitEvent(key);
		fireGeneralHitEvents(key, ch!=KeyEvent.CHAR_UNDEFINED);
	}

	/* (non-Javadoc)
	 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
	 */
	public void keyTyped(KeyEvent e) {
	}

	/**
	 * Adds a press event listener to the keyboard manager.
	 * @param keystroke The combination of keys to listen for.
	 * 					The format of this string is "([modifiers ](character|keyName)|All)"
	 * 						modifiers is in the format "(Ctrl[+Alt][+Shift]|Alt[+Shift]|Shift)"
	 * 						keyName is "(key|AllKeys|AllFnKeys)"
	 * 							key is what KeyEvent.getKeyText() returns.
	 * 						character is "(char|AllChars|AllNum|AllAlpha|AllAlphaNum|AllNonAlphaNum)";
	 * 							char is what KeyEvent.getKeyChar() returns.
	 * 					If the Shift modifier was used when typing a
	 * 					character it is removed from the modifers list.
	 * 					If a modifier key was the key that was pressed it's removed
	 * 					from the modifers list.
	 * @param lObj Listener object.
	 */
	public void addPressListener(String keystroke, ActionListener lObj){
		keystroke=getCanKeyStroke(keystroke);
		Vector v=(Vector)pressList.get(keystroke);
		if(v==null)
			pressList.put(keystroke, (v=new Vector()));
		// prevent doubles
		if(lObj!=null && v.indexOf(lObj)==-1)
			v.add(lObj);
	}
	
	/**
	 * Removes a press event listener.
	 * @param keystroke Combination of keys that the listener is listening to.
	 * @param lObj Listener object to remove.
	 */
	public void removePressListener(String keystroke, ActionListener lObj){
		keystroke=getCanKeyStroke(keystroke);
		Vector v=(Vector)pressList.get(keystroke);
		if(lObj==null || v==null) return;
		v.remove(lObj);
	}
	
	/**
	 * Returns the press event listeners for a particular keystroke.
	 * @param keystroke Combination of keys.
	 * @return Press listeners for that keystroke.
	 */
	public Vector getPressListeners(String keystroke){
		keystroke=getCanKeyStroke(keystroke);
		return (Vector)pressList.get(keystroke);
	}
	
	/**
	 * Returns all press listeners.
	 * @return All press listeners.
	 */
	public Hashtable getPressListeners(){
		return pressList;
	}
	
	/**
	 * Fires a press event for a particular keystroke.
	 * @param keystroke Keystroke to fire the event for.
	 */
	protected void firePressEvent(String keystroke){
		Vector v=getPressListeners(keystroke);
		if(v==null) return;
		int modifiers=0;
		String mods;
		if(keystroke.indexOf(' ')!=-1 && (mods=keystroke.substring(0, keystroke.indexOf(" "))).length()>0){
			if(mods.matches("Ctrl")) modifiers=modifiers | ActionEvent.CTRL_MASK;
			if(mods.matches("Shift")) modifiers=modifiers | ActionEvent.SHIFT_MASK;
			if(mods.matches("Alt")) modifiers=modifiers | ActionEvent.ALT_MASK;
		}
		ActionEvent e=new ActionEvent(this, ActionEvent.ACTION_PERFORMED, keystroke, modifiers);
		for(int i=0; i<v.size(); i++){
			if(v.get(i)==null){
				v.remove(i--);
			}else
				((ActionListener)v.get(i)).actionPerformed(e);
		}
	}

	/**
	 * Adds a release event listener to the keyboard manager.
	 * @param keystroke The combination of keys to listen for.
	 * 					The format of this string is "([modifiers ](character|keyName)|All)"
	 * 						modifiers is in the format "(Ctrl[+Alt][+Shift]|Alt[+Shift]|Shift)"
	 * 						keyName is "(key|AllKeys|AllFnKeys)"
	 * 							key is what KeyEvent.getKeyText() returns.
	 * 						character is "(char|AllChars|AllNum|AllAlpha|AllAlphaNum|AllNonAlphaNum)";
	 * 							char is what KeyEvent.getKeyChar() returns.
	 * 					If the Shift modifier was used when typing a
	 * 					character it is removed from the modifers list.
	 * 					If a modifier key was the key that was pressed it's removed
	 * 					from the modifers list.
	 * @param lObj Listener object.
	 */
	public void addReleaseListener(String keystroke, ActionListener lObj){
		keystroke=getCanKeyStroke(keystroke);
		Vector v=(Vector)releaseList.get(keystroke);
		if(v==null)
			releaseList.put(keystroke, (v=new Vector()));
		// prevent doubles
		if(lObj!=null && v.indexOf(lObj)==-1)
			v.add(lObj);
	}
	
	/**
	 * Removes a release event listener.
	 * @param keystroke Combination of keys that the listener is listening to.
	 * @param lObj Listener object to remove.
	 */
	public void removeReleaseListener(String keystroke, ActionListener lObj){
		keystroke=getCanKeyStroke(keystroke);
		Vector v=(Vector)releaseList.get(keystroke);
		if(lObj==null || v==null) return;
		v.remove(lObj);
	}
	
	/**
	 * Returns the release event listeners for a particular keystroke.
	 * @param keystroke Combination of keys.
	 * @return Press listeners for that keystroke.
	 */
	public Vector getReleaseListeners(String keystroke){
		keystroke=getCanKeyStroke(keystroke);
		return (Vector)releaseList.get(keystroke);
	}
	
	/**
	 * Returns all release listeners.
	 * @return All release listeners.
	 */
	public Hashtable getReleaseListeners(){
		return releaseList;
	}
	
	/**
	 * Fires a release event for a particular keystroke.
	 * @param keystroke Keystroke to fire the event for.
	 */
	protected void fireReleaseEvent(String keystroke){
		Vector v=getReleaseListeners(keystroke);
		if(v==null) return;
		int modifiers=0;
		String mods;
		if(keystroke.indexOf(' ')!=-1 && (mods=keystroke.substring(0, keystroke.indexOf(" "))).length()>0){
			if(mods.matches("Ctrl")) modifiers=modifiers | ActionEvent.CTRL_MASK;
			if(mods.matches("Shift")) modifiers=modifiers | ActionEvent.SHIFT_MASK;
			if(mods.matches("Alt")) modifiers=modifiers | ActionEvent.ALT_MASK;
		}
		ActionEvent e=new ActionEvent(this, ActionEvent.ACTION_PERFORMED, keystroke, modifiers);
		for(int i=0; i<v.size(); i++){
			if(v.get(i)==null){
				v.remove(i--);
			}else
				((ActionListener)v.get(i)).actionPerformed(e);
		}
	}

	/**
	 * Adds a hit event listener to the keyboard manager.
	 * @param keystroke The combination of keys to listen for.
	 * 					The format of this string is "([modifiers ](character|keyName)|All)"
	 * 						modifiers is in the format "(Ctrl[+Alt][+Shift]|Alt[+Shift]|Shift)"
	 * 						keyName is "(key|AllKeys|AllFnKeys)"
	 * 							key is what KeyEvent.getKeyText() returns.
	 * 						character is "(char|AllChars|AllNum|AllAlpha|AllAlphaNum|AllNonAlphaNum)";
	 * 							char is what KeyEvent.getKeyChar() returns.
	 * 					If the Shift modifier was used when typing a
	 * 					character it is removed from the modifers list.
	 * 					If a modifier key was the key that was pressed it's removed
	 * 					from the modifers list.
	 * @param lObj Listener object.
	 */
	public void addHitListener(String keystroke, ActionListener lObj){
		keystroke=getCanKeyStroke(keystroke);
		Vector v=(Vector)hitList.get(keystroke);
		if(v==null)
			hitList.put(keystroke, (v=new Vector()));
		// prevent doubles
		if(lObj!=null && v.indexOf(lObj)==-1)
			v.add(lObj);
	}
	
	/**
	 * Removes a hit event listener.
	 * @param keystroke Combination of keys that the listener is listening to.
	 * @param lObj Listener object to remove.
	 */
	public void removeHitListener(String keystroke, ActionListener lObj){
		keystroke=getCanKeyStroke(keystroke);
		Vector v=(Vector)hitList.get(keystroke);
		if(lObj==null || v==null) return;
		v.remove(lObj);
	}
	
	/**
	 * Returns the hit event listeners for a particular keystroke.
	 * @param keystroke Combination of keys.
	 * @return Press listeners for that keystroke.
	 */
	public Vector getHitListeners(String keystroke){
		return (Vector)hitList.get(getCanKeyStroke(keystroke));
	}
	
	/**
	 * Returns all hit listeners.
	 * @return All hit listeners.
	 */
	public Hashtable getHitListeners(){
		return hitList;
	}
	
	/**
	 * Fires a hit event for a particular keystroke.
	 * @param keystroke Keystroke to fire the event for.
	 */
	protected void fireHitEvent(String keystroke){
		Vector v=getHitListeners(keystroke);
		if(v==null) return;
		int modifiers=0;
		String mods;
		if(keystroke.indexOf(' ')!=-1 && (mods=keystroke.substring(0, keystroke.indexOf(" "))).length()>0){
			if(mods.matches("Ctrl")) modifiers=modifiers | ActionEvent.CTRL_MASK;
			if(mods.matches("Shift")) modifiers=modifiers | ActionEvent.SHIFT_MASK;
			if(mods.matches("Alt")) modifiers=modifiers | ActionEvent.ALT_MASK;
		}
		ActionEvent e=new ActionEvent(this, ActionEvent.ACTION_PERFORMED, keystroke, modifiers);
		for(int i=0; i<v.size(); i++){
			if(v.get(i)==null){
				v.remove(i--);
			}else
				((ActionListener)v.get(i)).actionPerformed(e);
		}
	}
	
	/**
	 * returns canonical form for the keystroke.
	 * @param keystroke Combination of keys.
	 * @return Canonical form of the keystroke.
	 */
	private String getCanKeyStroke(String keystroke){
		int sp=keystroke.indexOf(" ");
		if(keystroke.length()==1)
			keystroke=keystroke.toLowerCase();
		else if (sp!=-1 && keystroke.substring(sp+1).length()==1)
			keystroke=keystroke.substring(0, sp+1)+Character.toLowerCase(keystroke.charAt(sp+1));
		return keystroke;
	}
}
